/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2025 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system                           *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.20.0.2                                        *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------

Purpose : Thread safety functions.

Additional information:
  This module provides thread safety and interrupt safety for commonly
  used standard library functions. If standard library functions are
  used for which thread safety is not implemented yet, this file needs
  to be updated accordingly by adding thread safety for the functions
  to be used.
*/

#define _GNU_SOURCE

#include "RTOS.h"
#include <stdlib.h>
#include <stdio.h>
#include <dlfcn.h>
#include <stdarg.h>
#include <unistd.h>
#include <malloc.h>

/*********************************************************************
*
*       Configuration
*
**********************************************************************
*/
#if __GLIBC_PREREQ(2, 34)
  #define USE_MALLOC_HOOKS  (0)
#else
  #define USE_MALLOC_HOOKS  (1)
#endif

/*********************************************************************
*
*       Types, local
*
**********************************************************************
*/

#if (USE_MALLOC_HOOKS != 0)
typedef void*   OS_FUNC_MALLOC_HOOK(size_t, const void*);
typedef void*   OS_FUNC_REALLOC_HOOK(void*, size_t, const void*);
typedef void    OS_FUNC_FREE_HOOK(void*, const void*);
typedef void*   OS_FUNC_MEMALIGN_HOOK(size_t, size_t, const void*);
#else
typedef void*   OS_FUNC_MALLOC(size_t);
typedef void    OS_FUNC_FREE(void*);
typedef void*   OS_FUNC_CALLOC(size_t, size_t);
typedef void*   OS_FUNC_REALLOC(void*, size_t);
#endif

typedef int     OS_FUNC_PUTS(const char*);
typedef int     OS_FUNC_FPUTS(const char*, FILE*);
typedef ssize_t OS_FUNC_WRITE(int, const void*, size_t);
typedef size_t  OS_FUNC_FWRITE(const void* restrict, size_t, size_t, FILE* restrict);
typedef ssize_t OS_FUNC_READ(int, void*, size_t);

/*********************************************************************
*
*       Prototypes
*
**********************************************************************
*/
#if (USE_MALLOC_HOOKS != 0)
void OS_SIM_InitHooks(void);

static void* _MallocHook  (size_t Size, const void* Caller);
static void* _ReallocHook (void* Addr, size_t Size, const void* Caller);
static void  _FreeHook    (void* Addr, const void* Caller);
static void* _MemAlignHook(size_t Alignment, size_t Size, const void* Caller);
static void  _SaveHooks   (void);
static void  _RestoreHooks(void);
#endif

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/

static int _Initialized = 0;

#if (USE_MALLOC_HOOKS != 0)
static OS_FUNC_MALLOC_HOOK*   _OldMallocHook;
static OS_FUNC_REALLOC_HOOK*  _OldReallocHook;
static OS_FUNC_FREE_HOOK*     _OldFreeHook;
static OS_FUNC_MEMALIGN_HOOK* _OldMemAlignHook;
#else
static OS_FUNC_MALLOC*  pfMalloc  = NULL;
static OS_FUNC_FREE*    pfFree    = NULL;
static OS_FUNC_CALLOC*  pfCalloc  = NULL;
static OS_FUNC_REALLOC* pfRealloc = NULL;
#endif

static OS_FUNC_PUTS*    pPuts     = NULL;
static OS_FUNC_FPUTS*   pFputs    = NULL;
static OS_FUNC_WRITE*   pWrite    = NULL;
static OS_FUNC_FWRITE*  pFwrite   = NULL;
static OS_FUNC_READ*    pRead     = NULL;

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

/*********************************************************************
*
*       _InitThreadSafety()
*
*  Function description
*    Loads the addresses of the original libc functions which are used
*    by the wrapper functions in this file.
*
*  Additional information
*    Not all wrapper functions that provide thread safety do call
*    their counterpart but my call similar library functions.
*    For example, printf() calls vprintf() and therefore does not need
*    to load the address of printf() from the libc library.
*/
static void _InitThreadSafety(void) {
#if (USE_MALLOC_HOOKS == 0)
  pfMalloc  = (OS_FUNC_MALLOC*) dlsym(RTLD_NEXT, "malloc");
  pfFree    = (OS_FUNC_FREE*)   dlsym(RTLD_NEXT, "free");
  pfCalloc  = (OS_FUNC_CALLOC*) dlsym(RTLD_NEXT, "calloc");
  pfRealloc = (OS_FUNC_REALLOC*)dlsym(RTLD_NEXT, "realloc");
  if (pfMalloc == NULL) {
    fprintf(stderr, "[ERROR] Could not load run-time address of malloc(): %s\n", dlerror());
    exit(EXIT_FAILURE);
  }
  if (pfFree == NULL) {
    fprintf(stderr, "[ERROR] Could not load run-time address of free(): %s\n", dlerror());
    exit(EXIT_FAILURE);
  }
  if (pfCalloc == NULL) {
    fprintf(stderr, "[ERROR] Could not load run-time address of calloc(): %s\n", dlerror());
    exit(EXIT_FAILURE);
  }
  if (pfRealloc == NULL) {
    fprintf(stderr, "[ERROR] Could not load run-time address of realloc(): %s\n", dlerror());
    exit(EXIT_FAILURE);
  }
#endif
  pPuts   = (OS_FUNC_PUTS*)  dlsym(RTLD_NEXT, "puts");
  pFputs  = (OS_FUNC_FPUTS*) dlsym(RTLD_NEXT, "fputs");
  pWrite  = (OS_FUNC_WRITE*) dlsym(RTLD_NEXT, "write");
  pFwrite = (OS_FUNC_FWRITE*)dlsym(RTLD_NEXT, "fwrite");
  pRead   = (OS_FUNC_READ*)  dlsym(RTLD_NEXT, "read");
  if (pPuts == NULL) {
    fprintf(stderr, "[ERROR] Could not load run-time address of puts(): %s\n", dlerror());
    exit(EXIT_FAILURE);
  }
  if (pFputs == NULL) {
    fprintf(stderr, "[ERROR] Could not load run-time address of fputs(): %s\n", dlerror());
    exit(EXIT_FAILURE);
  }
  if (pWrite == NULL) {
    fprintf(stderr, "[ERROR] Could not load run-time address of write(): %s\n", dlerror());
    exit(EXIT_FAILURE);
  }
  if (pFwrite == NULL) {
    fprintf(stderr, "[ERROR] Could not load run-time address of fwrite(): %s\n", dlerror());
    exit(EXIT_FAILURE);
  }
  if (pRead == NULL) {
    fprintf(stderr, "[ERROR] Could not load run-time address of read(): %s\n", dlerror());
    exit(EXIT_FAILURE);
  }
  _Initialized = 1;
}

#if (USE_MALLOC_HOOKS != 0)
/*********************************************************************
*
*       _SaveHooks()
*/
static void _SaveHooks(void) {
  __malloc_hook   = _OldMallocHook;
  __realloc_hook  = _OldReallocHook;
  __free_hook     = _OldFreeHook;
  __memalign_hook = _OldMemAlignHook;
}

/*********************************************************************
*
*       _RestoreHooks()
*/
static void _RestoreHooks(void) {
  __malloc_hook   = _MallocHook;
  __realloc_hook  = _ReallocHook;
  __free_hook     = _FreeHook;
  __memalign_hook = _MemAlignHook;
}

/*********************************************************************
*
*       _MallocHook()
*/
static void* _MallocHook(size_t Size, const void* Caller) {
  void* Result;

  OS_USE_PARA(Caller);
  OS_ThreadSafe_Lock();
  _SaveHooks();
  Result = malloc(Size);
  _RestoreHooks();
  OS_ThreadSafe_Unlock();
  return Result;
}

/*********************************************************************
*
*       _ReallocHook()
*/
static void* _ReallocHook(void* Addr, size_t Size, const void* Caller) {
  void* Result;

  OS_USE_PARA(Caller);
  OS_ThreadSafe_Lock();
  _SaveHooks();
  Result = realloc(Addr, Size);
  _RestoreHooks();
  OS_ThreadSafe_Unlock();
  return Result;
}

/*********************************************************************
*
*       _FreeHook()
*/
static void _FreeHook(void* Addr, const void* Caller) {
  OS_USE_PARA(Caller);
  OS_ThreadSafe_Lock();
  _SaveHooks();
  free(Addr);
  _RestoreHooks();
  OS_ThreadSafe_Unlock();
}

/*********************************************************************
*
*       _MemAlignHook()
*/
static void* _MemAlignHook(size_t Alignment, size_t Size, const void* Caller) {
  void* Result;

  OS_USE_PARA(Caller);
  OS_ThreadSafe_Lock();
  _SaveHooks();
  Result = aligned_alloc(Alignment, Size);
  _RestoreHooks();
  OS_ThreadSafe_Unlock();
  return Result;
}
#endif

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_SIM_InitHook()
*/
#if (USE_MALLOC_HOOKS != 0)
void OS_SIM_InitHooks(void) {
  _OldMallocHook   = __malloc_hook;
  _OldReallocHook  = __realloc_hook;
  _OldFreeHook     = __free_hook;
  _OldMemAlignHook = __memalign_hook;
  __malloc_hook    = _MallocHook;
  __realloc_hook   = _ReallocHook;
  __free_hook      = _FreeHook;
  __memalign_hook  = _MemAlignHook;
}
#else
/*********************************************************************
*
*       malloc()
*/
void* malloc(size_t Size) {
  void* pAddress;

  if (_Initialized == 0) {
    _InitThreadSafety();
  }
  OS_ThreadSafe_Lock();
  pAddress = pfMalloc(Size);
  OS_ThreadSafe_Unlock();
  return pAddress;
}

/*********************************************************************
*
*       free()
*/
void free(void* pAddress) {
  if (_Initialized == 0) {
    _InitThreadSafety();
  }
  OS_ThreadSafe_Lock();
  pfFree(pAddress);
  OS_ThreadSafe_Unlock();
}

/*********************************************************************
*
*       calloc()
*/
void* calloc(size_t NumObjects, size_t Size) {
  void* pAddress;

  if (_Initialized == 0) {
    _InitThreadSafety();
  }
  OS_ThreadSafe_Lock();
  pAddress = pfCalloc(NumObjects, Size);
  OS_ThreadSafe_Unlock();
  return pAddress;
}

/*********************************************************************
*
*       realloc()
*/
void* realloc(void* pAddress, size_t Size) {
  void* pNewAddress;

  if (_Initialized == 0) {
    _InitThreadSafety();
  }
  OS_ThreadSafe_Lock();
  pNewAddress = pfRealloc(pAddress, Size);
  OS_ThreadSafe_Unlock();
  return pNewAddress;
}
#endif

/*********************************************************************
*
*       printf()
*/
int printf(const char* restrict pFormat, ...) {
  va_list Args;
  int     Result;

  va_start(Args, pFormat);
  OS_ThreadSafe_Lock();
  Result = vprintf(pFormat, Args);
  OS_ThreadSafe_Unlock();
  va_end(Args);
  return Result;
}

/*********************************************************************
*
*       fprintf()
*/
int fprintf(FILE* restrict pStream, const char* restrict pFormat, ...) {
  va_list Args;
  int     Result;

  va_start(Args, pFormat);
  OS_ThreadSafe_Lock();
  Result = vfprintf(pStream, pFormat, Args);
  OS_ThreadSafe_Unlock();
  va_end(Args);
  return Result;
}

/*********************************************************************
*
*       puts()
*/
int puts(const char* pStr) {
  int Result;

  if (_Initialized == 0) {
    _InitThreadSafety();
  }
  OS_ThreadSafe_Lock();
  Result = pPuts(pStr);
  OS_ThreadSafe_Unlock();
  return Result;
}

/*********************************************************************
*
*       fputs()
*/
int fputs(const char* restrict pStr, FILE* restrict Stream) {
  int Result;

  if (_Initialized == 0) {
    _InitThreadSafety();
  }
  OS_ThreadSafe_Lock();
  Result = pFputs(pStr, Stream);
  OS_ThreadSafe_Unlock();
  return Result;
}

/*********************************************************************
*
*       write()
*/
ssize_t write(int FileDescriptor, const void* pBuffer, size_t Count) {
  ssize_t Result;

  if (_Initialized == 0) {
    _InitThreadSafety();
  }
  OS_ThreadSafe_Lock();
  Result = pWrite(FileDescriptor, pBuffer, Count);
  OS_ThreadSafe_Unlock();
  return Result;
}

/*********************************************************************
*
*       fwrite()
*/
size_t fwrite(const void* restrict pData, size_t Size, size_t NumObjects, FILE* restrict Stream) {
  size_t Result;

  if (_Initialized == 0) {
    _InitThreadSafety();
  }
  OS_ThreadSafe_Lock();
  Result = pFwrite(pData, Size, NumObjects, Stream);
  OS_ThreadSafe_Unlock();
  return Result;
}

/*************************** End of file ****************************/
