/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2025 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system                           *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.20.0.2                                        *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
Purpose : Initializes and handles the hardware for embOS.
*/

#define _GNU_SOURCE
#include <stdlib.h>
#include <time.h>

#include "RTOS.h"
#include "UDPCOM.h"
#if ((OS_SUPPORT_TRACE_API != 0) && !(OS_64BIT))
#include "SEGGER_SYSVIEW_Sim.h"
#endif

/*********************************************************************
*
*       Defines
*
**********************************************************************
*/
#define NSEC_PER_SEC       (1000000000)
#define CYCLES_TO_NSEC(x)  ((x) * ((NSEC_PER_SEC) / (OS_TIMER_FREQ)))
#define NSEC_TO_CYCLES(x)  ((x) / ((NSEC_PER_SEC) / (OS_TIMER_FREQ)))
#define SEC_TO_CYCLES(x)   ((x) * (OS_TIMER_FREQ))

/*********************************************************************
*
*       System tick settings
*/
//
// The current implementation of the RTOSInit.c allows to specify the
// timer frequency to be used for the embOS simulation. In order to
// change the timer frequency, only OS_TIMER_FREQ needs to be changed.
// Since clock_nanosleep() supports nanosecond resolution, frequencies
// up to 1 GHz can be used for OS_TIMER_FREQ. Changing OS_TIMER_FREQ
// also affects the cycle resolution. By default, OS_TIMER_FREQ is set
// to 1000000, i.e. 1 MHz, meaning that one cycle equals one microsecond.
//
//
#define OS_TIMER_FREQ       (1000000)  // 1 MHz
#define OS_TICK_FREQ        (1000)     // 1000 ticks per second
#define OS_INT_FREQ         (OS_TICK_FREQ)
#define OS_CYCLES_PER_TICK  (OS_TIMER_FREQ / OS_TICK_FREQ)

#if (OS_TIMER_FREQ > NSEC_PER_SEC)
  #error "Timer frequency values greater than 1 GHz are not supported."
#endif

/*********************************************************************
*
*       embOSView settings
*/
#ifndef   OS_VIEW_IFSELECT
#if (OS_64BIT == 1)
  #define OS_VIEW_IFSELECT  OS_VIEW_DISABLED  // embOSView communication currently not supported for devices with 64-bit address range
#else
  #define OS_VIEW_IFSELECT  OS_VIEW_IF_ETHERNET
#endif
#endif

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/
static struct timespec _LastSystickTimestamp;
static struct timespec _NextSysTickTimestamp;

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

/*********************************************************************
*
*       _GetCycles()
*
*  Function description
*    If pTimespec is NULL, the current time stamp in cycles is
*    returned. Else, the time stamp pointed to by pTimespec is
*    converted to cycles and returned.
*/
static OS_U64 _GetCycles(const struct timespec* pTimespec) {
  struct timespec Timespec;
  OS_U64          Timestamp;

  if (pTimespec == NULL) {
    clock_gettime(CLOCK_MONOTONIC, &Timespec);
    pTimespec = &Timespec;
  }
  Timestamp  =  SEC_TO_CYCLES(pTimespec->tv_sec);
  Timestamp += NSEC_TO_CYCLES(pTimespec->tv_nsec);
  return Timestamp;
}

/*********************************************************************
*
*       _AddTime()
*
*  Function description
*    Adds the specified nanoseconds to the timespec structure and
*    ensures that the value of the tv_nsec member doesn't exceed
*    1 billion.
*
*  Parameters
*    pTimespec: Pointer to the timespec structure to adjust.
*    ns:        The nanoseconds to add to the timespec structure.
*/
static void _AddTime(struct timespec* pTimespec, int ns) {
  if (pTimespec->tv_nsec >= (NSEC_PER_SEC - ns)) {
    pTimespec->tv_nsec -= NSEC_PER_SEC - ns;
    pTimespec->tv_sec  += 1;
  } else {
    pTimespec->tv_nsec += ns;
  }
}

/*********************************************************************
*
*       _SystemTickController()
*
*  Function description
*    This function controls the occurrence of system tick interrupts.
*/
static int _SystemTickController(void) {
  static int InitOnce = 1;
  int        Result;

  if (InitOnce != 0) {
    InitOnce = 0;
#if ((OS_SUPPORT_TRACE_API != 0) && !(OS_64BIT))
    SEGGER_SYSVIEW_X_SetISRName("Tick ISR");
#endif
    //
    // Get the initial time stamp.
    //
    clock_gettime(CLOCK_MONOTONIC, &_NextSysTickTimestamp);
  }
  //
  // Calculate the absolute time stamp for the next system tick.
  //
  _AddTime(&_NextSysTickTimestamp, CYCLES_TO_NSEC(OS_CYCLES_PER_TICK));
  //
  // Delay until the newly calculated time stamp.
  //
  do {
    Result = clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME, &_NextSysTickTimestamp, NULL);
  } while (Result != 0);
  return 1;  // Execute the ISR.
}

/*********************************************************************
*
*       _SystemTickHandler()
*/
static void _SystemTickHandler(void) {
  OS_INT_Enter();
  OS_TICK_Handle();
  //
  // Save the time stamp of the last system tick that was handled.
  // This information is used to calculate the elapsed cycles.
  //
  _LastSystickTimestamp = _NextSysTickTimestamp;
  OS_INT_Leave();
}

/*********************************************************************
*
*       _OS_GetHWTimerCycles()
*
*  Function description
*    Returns the current hardware timer count value.
*
*  Return value
*    Current timer count value.
*/
static unsigned int _OS_GetHWTimerCycles(void) {
  OS_U64 Cycles;

  //
  // Convert the timespec structure of the last system
  // tick to cycles and use it to calculate the elapsed
  // cycles since the last system tick.
  //
  Cycles = _GetCycles(&_LastSystickTimestamp);
  Cycles = _GetCycles(NULL) - Cycles;
  return (unsigned int)Cycles;
}

/*********************************************************************
*
*       _OS_GetHWTimer_IntPending()
*
*  Function description
*    Returns if the hardware timer interrupt pending flag is set.
*
*  Return value
*    == 0: Interrupt pending flag not set.
*    != 0: Interrupt pending flag set.
*/
static unsigned int _OS_GetHWTimer_IntPending(void) {
  return 0;
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_InitHW()
*
*  Function description
*    Initialize the hardware required for embOS to run.
*/
void OS_InitHW(void) {
  OS_INT_IncDI();
  OS_SIM_CreateISRThreadEx(_SystemTickHandler, _SystemTickController, "System Tick ISR");
  {
    OS_SYSTIMER_CONFIG SysTimerConfig = {OS_TIMER_FREQ, OS_INT_FREQ, OS_TIMER_UPCOUNTING, _OS_GetHWTimerCycles, _OS_GetHWTimer_IntPending};
    OS_TIME_ConfigSysTimer(&SysTimerConfig);
  }
  //
  // Initialize communication for embOSView
  //
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_ETHERNET)
  UDP_Process_Init();
#endif
  //
  // Configure and initialize SEGGER SystemView
  //
#if ((OS_SUPPORT_TRACE_API != 0) && !(OS_64BIT))
  SEGGER_SYSVIEW_Conf();
#endif
  OS_INT_DecRI();
}

/*********************************************************************
*
*       Optional communication with embOSView
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_COM_Send1()
*
*  Function description
*    Sends one character.
*/
void OS_COM_Send1(OS_U8 c) {
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_ETHERNET)
  UDP_Process_Send1(c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_DISABLED)
  OS_USE_PARA(c);          // Avoid compiler warning
  OS_COM_ClearTxActive();  // Let embOS know that Tx is not busy
#endif
}

/*************************** End of file ****************************/
